-------------------------------------------------------------------------------
-- Package Calendar_Utilities (body)                                         --
-- (C) Copyright 2001 ADALOG                                                 --
-- Author: J-P. Rosen                                                        --
--                                                                           --
-- Rights to use, distribute or modify this package in any way is hereby     --
-- granted, provided this header is kept unchanged in all versions and the   --
-- associated documentation file is distributed unchanged. Additionnal       --
-- headers or documentation may be added.                                    --
-- All modifications must be properly marked as not originating from Adalog. --
-- If you make a valuable addition, please keep us informed by sending a     --
-- message to rosen@adalog.fr                                                --
--                                                                           --
-- ADALOG is providing training, consultancy and expertise in Ada and        --
-- related software engineering techniques. For more info about our services:--
-- ADALOG                   Tel: +33 1 41 24 31 40                           --
-- 19-21 rue du 8 mai 1945  Fax: +33 1 41 24 07 36                           --
-- 94110 ARCUEIL            E-m: info@adalog.fr                              --
-- FRANCE                   URL: http://www.adalog.fr                        --
--                                                                           --
-- This package is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY  --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                      --
-------------------------------------------------------------------------------

with Ada.Integer_Text_IO; use Ada.Integer_Text_IO;
with Ada.Text_IO;
with Ada.Strings.Fixed;
with Ada.Strings.Maps;
package body Calendar_Utilities is

   -----------
   -- Image --
   -----------

   function Image (Item : Time; Format : String) return String is
      Buffer : String (1..Max_Image_Length);
      Length : Natural := 0;

      procedure Append (Item : Integer; Width : Positive) is
      begin
         Put (Buffer (Length+1..Length+Width), Item);
         for I in Length+1 .. Length+Width loop
            if Buffer (I) = ' ' then
               Buffer (I) := '0';
            end if;
         end loop;
         Length := Length + Width;
      end Append;
      procedure Append (Item : Character) is
      begin
         Length := Length + 1;
         Buffer (Length) := Item;
      end Append;

      Special : Boolean := False;
      Temp_D  : Natural := Natural (Seconds(Item));
   begin
      if Duration(Temp_D) > Seconds (Item) then
         -- Truncate, don't round
         Temp_D := Temp_D - 1;
      end if;

      for I in Format'Range loop
         if Special then
            case Format(I) is
               when 'Y' =>
                  Append (Year (Item), 4);
               when 'M' =>
                  Append (Month (Item), 2);
               when 'D' =>
                  Append (Day (Item), 2);
               when 'h' =>
                  Append (Temp_D / 3600, 2);
               when 'm' =>
                  Append ((Temp_D rem 3600) / 60, 2);
               when 's' =>
                  Append (Temp_D rem 60, 2);
               when 'c' =>
                  Append (Natural((Seconds(Item) - Duration(Temp_D)) * 100), 2);
               when others =>
                  Append (Format(I));
            end case;
            Special := False;
         elsif Format(I) = '%' then
            Special := True;
         else
            Append (Format(I));
         end if;
      end loop;

      return Buffer (1..Length);
   end Image;

   ---------
   -- Max --
   ---------

   function Max (Left, Right : Time) return Time is
   begin
      if Left > Right then
         return Left;
      else
         return Right;
      end if;
   end Max;

   ---------
   -- Min --
   ---------

   function Min (Left, Right : Time) return Time is
   begin
      if Left < Right then
         return Left;
      else
         return Right;
      end if;
   end Min;

   -----------
   -- Value --
   -----------

   subtype Hour_Range   is Integer range 0..23;
   subtype Minsec_Range is Integer range 0..59;
   use Ada.Strings.Maps;
   colon_to_semi_Mapping : constant Character_Mapping := To_Mapping (":", ";");

   function Value (Item : String; Format : String) return Time is
      use Ada.Text_IO, Ada.Strings.Fixed;
      -- Translate ':' to ';', otherwise Get interprets it as the alternative
      -- base symbol :-(
      Source : constant String (1..Item'Length) :=
        Translate (Item, Colon_To_Semi_Mapping);
      Model : constant String (1..Format'Length) :=
        Translate (Format, Colon_To_Semi_Mapping);

      Now  : constant Time := Clock;
      Temp : Natural;
      First: Natural := Source'First;
      Last : Natural;
      Special : Boolean := False;

      -- We use Natural here rather than Day_Number, Month_Number, Year_Number
      -- because we need some flexibility with years... Anyway, Time_Of will
      -- check the validity of the data
      D : Natural  := Day     (Now);
      M : Natural  := Month   (Now);
      Y : Natural  := Year    (Now);
      S : Duration := 0.0;

   begin
      begin
         for I in Model'Range loop
            if Special then
               case Model(I) is
                  when 'Y' =>
                     Get (Source(First..Integer'Min(First+3, Source'Last)), Y, Last);
                  when 'M' =>
                     Get (Source(First..Integer'Min(First+1, Source'Last)), M, Last);
                  when 'D' =>
                     Get (Source(First..Integer'Min(First+1, Source'Last)), D, Last);
                  when 'h' =>
                     Get (Source(First..Integer'Min(First+1, Source'Last)), Hour_Range(Temp), Last);
                     S := S + Duration (Temp*3600);
                  when 'm' =>
                     Get (Source(First..Integer'Min(First+1, Source'Last)), Minsec_Range(Temp), Last);
                     S := S + Duration (Temp*60);
                  when 's' =>
                     Get (Source(First..Integer'Min(First+1, Source'Last)), Minsec_Range(Temp), Last);
                     S := S + Duration (Temp);
                  when 'c' =>
                     Get (Source (First..Integer'Min(First+1, Source'Last)), Temp, Last);
                     S := S + Duration(Temp)/100;
                  when others =>
                     if Model (I) /= Source (First) then
                        raise Time_Error;
                     else
                        Last := First;
                     end if;
               end case;
               Special := False;
               First := Last + 1;
            elsif Model(I) = '%' then
               Special := True;
            elsif First > Source'Last then
               exit;
            else
               if Model (I) /= Source (First) then
                  raise Time_Error;
               else
                  First := First + 1;
               end if;
            end if;
         end loop;
      exception
         when End_Error =>
            null;
      end;

      if Y >= 80 and Y <= 99 then
         Y := Y + 1900;
      elsif Y <= 79 then
         Y := Y + 2000;
      end if;

      return Time_Of (Y, M, D, S);
   exception
      when Data_Error | Constraint_Error =>
         raise Time_Error;
   end Value;

end Calendar_Utilities;
