-------------------------------------------------------------------------------
-- Package OS_Services_TCL (spec)                                           --
-- (C) Copyright 2000 ADALOG                                                 --
-- Author: J-P. Rosen                                                        --
--                                                                           --
-- Rights to use, distribute or modify this package in any way is hereby     --
-- granted, provided this header is kept unchanged in all versions and the   --
-- associated documentation file is distributed unchanged. Additionnal       --
-- headers or documentation may be added.                                    --
-- All modifications must be properly marked as not originating from Adalog. --
-- If you make a valuable addition, please keep us informed by sending a     --
-- message to rosen.adalog@wanadoo.fr                                        --
--                                                                           --
-- ADALOG is providing training, consultancy and expertise in Ada and        --
-- related software engineering techniques. For more info about our services:--
-- ADALOG                   Tel: +33 1 41 24 31 40                           --
-- 19-21 rue du 8 mai 1945  Fax: +33 1 41 24 07 36                           --
-- 94110 ARCUEIL            E-m: rosen.adalog@wanadoo.fr                     --
--                          URL: http://pro.wanadoo.fr/adalog                --
--                                                                           --
-- This package is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY  --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                      --
-------------------------------------------------------------------------------
with  -- Standard Ada units
  Ada.Calendar,
  Ada.Command_Line,

  -- For private part only:
  Ada.Strings.Unbounded,
  Ada.Finalization;
package OS_Services_TCL is
   pragma Elaborate_Body;

   -- Properties of this implementation:
   -- ----------------------------------
   --
   -- Supported OSes:
   --    DOS, NT, OS_2 (untested), Unix, MacOS (untested)
   --
   -- Tasking:
   --    Operations in this package are NOT reentrant, unless specified.
   --
   -- Pattern pre-compilation:
   --    Pre-compilation is NOT effective.
   --
   -- Unimplemented properties:
   --    None


   -------------------------
   --                     --
   -- General OS services --
   --                     --
   -------------------------

   -- Exception raised if any of the provided services fails
   -- Exception_Message is set to a description of the reason
   -- for the failure
   Service_Error : exception;

   -- Do we (by default) ignore case in file name comparisons?
   -- Initialized in the body according to the OS, can
   -- be changed at any time.
   Default_Ignore_Case : Boolean;

   -- Which OS are we running under ?
   -- Subtype Dos_OS refer to OSes with (more or less) Dos naming system.
   type OS_Kind is (Unknown, Dos, NT, OS_2, Unix, MacOS); -- Unsupported: VMS
   subtype Dos_OS is OS_Kind range Dos .. OS_2;
   function Operating_System return OS_Kind;

   -- Get the value of an environment variable :
   function Get_Environment (Name : String) return String;

   -- Get (full path of) executable
   function Executable_Path return String;

   -- Execute OS command directly:
   procedure OS_Execute (Command  : in     String;
                         Result   :    out Ada.Command_Line.Exit_Status;
                         Blocking : in     Boolean := True);

   -- Execute OS command in a sub-shell:
   -- (under DOS, it is the only way to execute built-in commands)
   procedure Shell_Execute (Command  : in     String;
                            Result   :    out Ada.Command_Line.Exit_Status;
                            Blocking : in     Boolean := True);


   --------------------------------------------
   --                                        --
   -- File name pattern matching utilitities --
   -- (Glob-style matching)                  --
   --                                        --
   --------------------------------------------

   -- Check if "Name" matches "Pattern"
   function Match (Name : String; Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Boolean;

   -- On some implementations, it is possible to pre-compile a pattern. On these systems, using
   -- the following utilities may be more efficient if the same pattern is used more than once.
   -- On other implementations, the following utilities will work, but won't be more efficient.
   -- For this implementation (Tcl based): Pre-compilation is NOT effective.
   type Compiled_Pattern is private;
   function Compile (Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Compiled_Pattern;
   function Match   (Name : String; Pattern : Compiled_Pattern) return Boolean;


   ---------------------------------------
   --                                   --
   -- File name and properties services --
   --                                   --
   ---------------------------------------

   -- Catenate path and name with a directory separator:
   function Join (Path : String; Name : String) return String; --Reentrant

   type String_Property  is private;
   function "+" (Left, Right : String_Property) return String_Property;

   --  Given:   X := "subdir\name.ext", current dir = c:\dir

   -- Basic values:
   -- Expanded and As_File modify the formatting of the result.
   -- Device, Path, Name, Extension, Version extract the corresponding field, or
   -- return "" if the field is not present in the provided name.
   -- The input name can use any directory separator on any system; however, the
   -- result always uses the separator that is appropriate to the OS.
   -- If more than one field is desired, use "+".
   -- It is guaranteed that the various parts can be concatenated with "&" to form a
   -- legal name.
   Expanded  : constant String_Property; -- Expand the path to form an absolute name
   As_File   : constant String_Property; -- Consider the input string as a file name, even if there is a trailing separator
   Device    : constant String_Property; -- "c:"
   Path      : constant String_Property; -- "subdir\". If Expanded, "\dir\subdir\".
   Name      : constant String_Property; -- "name"
   Extension : constant String_Property; -- ".ext"
   Version   : constant String_Property; -- ";1"

   -- Native_Form is a String_Property that returns all the fields in the original string.
   -- It is not equivalent to doing nothing: all separators have been changed to the value
   -- which is correct for the OS - hence the name.
   Native_Form : constant String_Property; -- = Device + Path + Name + Extension + Version

   type Boolean_Property is (Exists, Is_Directory, Is_Regular_File, Is_Owned, Is_Readable, Is_Writable, Is_Executable);
   type Time_Property    is (Access_Time, Change_Time);
   type Long_Property    is (Size);
   type Enum_Property    is (File_Kind, Path_Kind);

   type Enum_Values is
     (File,     Directory, CharacterSpecial, BlockSpecial, Fifo, Link, Socket, -- Values for File_Kind
      Absolute, Relative,  VolumeRelative);                                    -- Values for Path_Kind
   subtype File_Kind_Enums    is Enum_Values range File..Socket;
   subtype Path_Kind_Enums    is Enum_Values range Absolute..VolumeRelative;

   type File_Properties (Kind : File_Kind_Enums := File) is
      record
         Name : Ada.Strings.Unbounded.Unbounded_String;
         Path : Ada.Strings.Unbounded.Unbounded_String;
         case Kind is
         when File | Directory =>
            Access_Time : Ada.Calendar.Time;
            Change_Time : Ada.Calendar.Time;
            Size        : Long_Integer;
         when others =>
            null;
         end case;
      end record;

   function Property (Query : String_Property;  Name : String; Default : String := "") return String;
   function Property (Query : Boolean_Property; Name : String) return Boolean;
   function Property (Query : Time_Property;    Name : String) return Ada.Calendar.Time;
   function Property (Query : Long_Property;    Name : String) return Long_Integer;
   function Property (Query : Enum_Property;    Name : String) return Enum_Values;

   function Properties (Name : String) return File_Properties;


   -----------------------------------
   --                               --
   --  Directory iterator functions --
   --                               --
   -----------------------------------

   type File_List is private;  -- Initialized to the "empy list" state
   function Build_List
      (Pattern : in  String  := "*";
       Sort    : in  Boolean := False) return File_List;
   function Length (List : File_List) return Natural;

   function "+" (Left, Right : File_List) return File_List;
   -- Merges two lists. If BOTH are sorted, then the result is sorted and duplicates are removed.

   function To_String (List : File_List; Separator : Character := ' ') return String;
   -- Gives a string representation of the list.
   -- File names are in absolute form, and separated with a Separator character
   -- File names that contain a Separator are in quotes.

   generic
      with function Is_Selected (List : File_List; Index : Positive) return Boolean;
   function Generic_Filter (List : File_List) return File_List;
   -- Keeps in the resulting list only the entries for which "Is_Selected" returned True

   subtype Iter_Enum_Property is Enum_Property range File_Kind..File_Kind;
   function Property (Query : String_Property;    List : File_List; Index : Positive) return String;
   function Property (Query : Boolean_Property;   List : File_List; Index : Positive) return Boolean;
   function Property (Query : Time_Property;      List : File_List; Index : Positive) return Ada.Calendar.Time;
   function Property (Query : Long_Property;      List : File_List; Index : Positive) return Long_Integer;
   function Property (Query : Iter_Enum_Property; List : File_List; Index : Positive) return Enum_Values;

   function Properties (List : File_List; Index : Positive) return File_Properties;


   -----------------------------------
   --                               --
   -- File and directory operations --
   --                               --
   -----------------------------------

   -- Operations whose first parameter is "Pattern" operate on all
   -- files that match "Pattern". Copy_File and Delete_File treat
   -- directories as normal files, i.e. they ARE deleted (even if non
   -- empty) or fully copied (including -recursively- their content).

   procedure Copy_File   (Pattern : String; To : String); --Reentrant
   procedure Move_File   (Pattern : String; To : String);
   procedure Rename_File (File    : String; To : String); --Reentrant
   procedure Delete_File (Pattern : String);              --Reentrant

   procedure Create_Dir (Name : String); --Reentrant
   procedure Change_Dir (To : String);
   function  Current_Dir return String;

private
   -- Note on naming conventions:
   -- Since the visible part of this package has several implementations, we used what seemed to us a
   -- "natural" convention to designate the various parts of a file name.
   -- In the body (and in the private part), the same fields are named after the convention
   -- used in the Tcl "file" command, except that we use "kind" instead of "type" for obvious
   -- reasons ;-)

   type String_Property_Flag is (Expanded_Flag, As_File_Flag,
                                 Device_Flag, Path_Flag, Name_Flag, Ext_Flag, Version_Flag);
   type String_Property is array (String_Property_Flag) of Boolean;

   -- Basic values:
   Expanded    : constant String_Property := (Expanded_Flag => True, others => False);
   As_File     : constant String_Property := (As_File_Flag  => True, others => False);
   Device      : constant String_Property := (Device_Flag   => True, others => False);
   Path        : constant String_Property := (Path_Flag     => True, others => False);
   Name        : constant String_Property := (Name_Flag     => True, others => False);
   Extension   : constant String_Property := (Ext_Flag      => True, others => False);
   Version     : constant String_Property := (Version_Flag  => True, others => False);
   Native_Form : constant String_Property := (Expanded_Flag | As_File_Flag => False,
                                              others                       => True);

   -- This type is like File_Info, but without discriminants and
   -- with additional fields for use inside this package.
   -- Since accessing properties might be costly over a network, we make a lazy evaluation of
   -- some fields; the Initialized discriminant tells whether the values have been evaluated.
   type Inner_Info (Initialized : Boolean := False) is
      record
         Tail    : Ada.Strings.Unbounded.Unbounded_String;
         Dirname : Ada.Strings.Unbounded.Unbounded_String;
         case Initialized is
         when False =>
            null;
         when True =>
            Kind  : File_Kind_Enums := File;
            Atime : Ada.Calendar.Time;
            Mtime : Ada.Calendar.Time;
            Size  : Long_Integer;
         end case;
      end record;
   type Info_Table is array (Positive range <>) of Inner_Info;
   type Info_List (Length : Natural) is
      record
         Ref_Count : Natural := 0;
         Sorted    : Boolean;
         Infos     : Info_Table (1..Length);
      end record;
   type Info_List_Access is access Info_List;

   type File_List is new Ada.Finalization.Controlled with
      record
         Content : Info_List_Access;
      end record;
   -- Nothing to do for Initialize.
   procedure Adjust     (Object : in out File_List);
   procedure Finalize   (Object : in out File_List);

   -- Precompilation is not available in Tcl for Tcl_StringMatch
   -- (and Tcl_RegExpMatch is not what we want).
   -- Therefore, we keep the "precompiled" pattern as the original string.
   type Compiled_Pattern is
     record
        Pattern     : Ada.Strings.Unbounded.Unbounded_String;
        Ignore_Case : Boolean;
     end record;
end OS_Services_TCL;
